
/******************************************************************************
* MODULE     : lexer.cc
* DESCRIPTION: Lexical analysis of mathemagix files
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <lexer.hh>
namespace mmx {
#include <tokens.hh>

int   lex_origin;
char* lex_string;
int   lex_length;
int   lex_start;
int   lex_prev;
int   lex_pos;
bool  lex_accolade_mode;

#define org   lex_origin
#define s     lex_string
#define n     lex_length
#define start lex_start
#define prev  lex_prev
#define pos   lex_pos

static string
make_keyword () {
  string r (pos- start+ 1);
  int i, j;
  for (i=0, j=start; j<pos; i++, j++)
    r[i]= s[j];
  r[i]= '!';
  return r;
}

#define produce(code) { \
  parse_lval= generic (symbol (string (s+ start, pos- start), \
			    org+ start, org+ pos)); \
  return code; }

#define produce_keyword(code) { \
  parse_lval= generic (symbol (make_keyword(), org+ start, org+ pos)); \
  return code; }

#define test(c,code) \
  if ((pos<n) && (s[pos]==c)) { pos++; produce (code); }

#define keyword(c,k,code) \
  if ((s[pos]==c) && test_keyword(k)) produce_keyword (code);

#define is_alpha(c) \
  ((((c)>='0') && ((c)<='9')) || \
   (((c)>='A') && ((c)<='Z')) || \
   (((c)>='a') && ((c)<='z')) || \
   (((c)=='_')))

bool
test_keyword (char* k) {
  pos= start;
  while ((*k)!='\0') {
    if (pos>=n) return false;
    if (s[pos]!=(*k)) return false;
    pos++;
    k++;
  }
  if (pos>=n) return true;
  if (is_alpha (s[pos])) return false;
  return true;
}

int
lex () {
  if (lex_accolade_mode) {
    parse_lval= generic (symbol ("", org+ pos, org+ pos));
    lex_accolade_mode= false;
    return ';';
  }

  if (pos == n) {
    s= NULL;
    parse_lval= generic ();
    return 0;
  }

  prev= pos;
  char c= s[pos];
  while ((c<=' ') || (c>'~')) {
    pos++;
    if (pos == n) {
      s= NULL;
      parse_lval= generic ();
      return 0;
    }
    c= s[pos];
  }

  start= pos++;
  switch (c) {
  case '!':
    test ('=', NOT_EQUAL);
    if ((pos<n) && (s[pos]=='<')) {
      pos++;
      test ('=', NOT_LEQ);
      produce (NOT_LESS);
    }
    if ((pos<n) && (s[pos]=='>')) {
      pos++;
      test ('=', NOT_GEQ);
      produce (NOT_GREATER);
    }
    produce (NOT);
  case '\042':
    while (pos<n) {
      if ((s[pos]=='\\') && ((pos+2)<n)) { pos+=2; continue; }
      if (s[pos]=='\042') { pos++; break; }
      pos++;
    }
    produce (STRING);
  case '#':
    produce (SIZE);
  case '$':
    produce (SIGNATURE);
  case '%':
    produce (THIS);
  case '&':
    test ('&', AND);
    produce (c);
  case '\047':
    produce (PRIME);
  case '(':
  case ')':
    produce (c);
  case '*':
    test ('=', TIMES_ASSIGN);
    produce (TIMES);
  case '+':
    test ('=', PLUS_ASSIGN);
    test ('+', INC);
    produce (PLUS);
  case ',':
    produce (c);
  case '-':
    test ('=', MINUS_ASSIGN);
    test ('-', DEC);
    test ('>', INTO);
    produce (MINUS);
  case '.':
    test ('.', RANGE);
    produce (ACCESS);
  case '/':
    if (s[pos]=='/') {
      pos++;
      while ((pos<n) && (s[pos]!='\n')) pos++;
      if (pos<n) pos++;
      return lex ();
    }
    if (s[pos]=='{') {
      int level= 1;
      pos++;
      while ((pos+1<n) && (level>0)) {
	if ((s[pos]=='/') && (s[pos+1]=='{')) { level++; pos+=2; continue; }
	if ((s[pos]=='}') && (s[pos+1]=='/')) { level--; pos+=2; continue; }
	pos++;
      }
      if (level>0) pos=n;
      return lex ();
    }
    test ('=', OVER_ASSIGN);
    produce (OVER);
  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
  case '7':
  case '8':
  case '9':
    while ((pos<n) && (s[pos]>='0') && (s[pos]<='9')) pos++;
    if ((pos<n+1) && (s[pos]=='.') && (s[pos+1]>='0') && (s[pos+1]<='9')) {
      pos+=2;
      while ((pos<n) && (s[pos]>='0') && (s[pos]<='9')) pos++;
      if ((pos<n+1) && ((s[pos]=='e') || (s[pos]=='E'))) {
	int p= ((pos<n+2) && (s[pos]=='-') ? 2: 1);
	if ((s[pos+p]>='0') && (s[pos+p]<='9')) {
	  pos+=p;
	  while ((pos<n) && (s[pos]>='0') && (s[pos]<='9')) pos++;
	}
      }
    }
    produce (NUMBER);
  case ':':
    test ('=', ASSIGN);
    test ('>', CAST);
    produce (TYPE);
  case ';':
    produce (c);
  case '<':
    test ('<', LEFT_FLUX);
    if ((pos+1<n) && (s[pos]=='=') && (s[pos+1]=='>')) {
      pos += 2;
      produce (EQUIVALENT);
    }
    test ('=', LEQ);
    produce (LESS);
  case '=':
    test ('=', DEFINE);
    test ('>', IMPLIES);
    produce (EQUAL);
  case '>':
    test ('>', RIGHT_FLUX);
    test ('=', GEQ);
    produce (GREATER);
  case '?':
    produce (c);
  case '@':
    test ('+', OPLUS);
    test ('-', OMINUS);
    test ('*', OTIMES);
    test ('/', OOVER);
    produce (COMPOSE);
  case 'A':
  case 'B':
  case 'C':
  case 'D':
  case 'E':
  case 'F':
  case 'G':
  case 'H':
  case 'I':
  case 'J':
  case 'K':
  case 'L':
  case 'M':
  case 'N':
  case 'O':
  case 'P':
  case 'Q':
  case 'R':
  case 'S':
  case 'T':
  case 'U':
  case 'V':
  case 'W':
  case 'X':
  case 'Y':
  case 'Z':
    goto identifier;
  case '[':
  case '\\':
  case ']':
    produce (c);
  case '^':
    test ('^', XOR);
    produce (POWER);
  case '_':
    goto identifier;
  case '`':
    produce (BACKPRIME);
  case 'a':
    if (pos<n) {
      keyword ('b', "abstract", ABSTRACT);
      keyword ('d', "add", ADD);
      keyword ('n', "and", SEQAND);
    }
    goto identifier;
  case 'b':
    if (pos<n) keyword ('r', "break", BREAK);
    goto identifier;
  case 'c':
    if ((pos+2)<n) {
      if (s[pos]=='a') {
	pos++;
	keyword ('s', "case", CASE);
	keyword ('t', "catch", CATCH);
      }
      if ((s[pos]=='o') && (s[pos+1]=='n')) {
	pos += 2;
	keyword ('c', "concrete", CONCRETE);
	keyword ('s', "constant", CONSTANT);
	keyword ('t', "continue", CONTINUE);
      }
    }
    goto identifier;
  case 'd':
    if (pos<n) {
      keyword ('i', "direct", DIRECT);
      keyword ('o', "do", DO);
    }
    goto identifier;
  case 'e':
    if ((pos+1)<n) {
      keyword ('l', "else", ELSE);
      keyword ('v', "evolutive", EVOLUTIVE);
      if (s[pos]=='x') {
	pos++;
	keyword ('i', "exists", EXISTS);
	keyword ('p', "export", EXPORT);
	if (((pos+2)<n) && (s[pos]=='t') && (s[pos+1]=='e')) {
	  pos += 2;
	  keyword ('n', "extend", EXTEND);
	  keyword ('r', "extern", EXTERN);
	}
      }
    }
    goto identifier;
  case 'f':
    if (pos<n) {
      if ((pos+1<n) && (s[pos]=='o') && (s[pos+1]=='r')) {
	if ((pos+2<n) && (s[pos+2]=='a')) {
	  keyword ('o', "forall", FORALL);
	}
	else if ((pos+2<n) && (s[pos+2]=='e')) {
	  keyword ('o', "foreign", FOREIGN);
	}
	else {
	  keyword ('o', "for", FOR);
	}
      }
      keyword ('r', "from", FROM);
    }
    goto identifier;
  case 'g':
    if (pos<n) keyword ('e', "generate", GENERATE);
    goto identifier;
  case 'h':
    if (pos<n) {
      keyword ('a', "has", HAS);
      keyword ('i', "hidden", HIDDEN);
      keyword ('o', "holds", HOLDS);
    }
    goto identifier;
  case 'i':
    if (pos<n) {
      keyword ('f', "if", IF);
      keyword ('m', "import", IMPORT);
      if (s[pos]=='n') {
	pos++;
	if ((pos >= n) || (!is_alpha (s[pos]))) produce (IN);
	keyword ('d', "indirect", INDIRECT);
	keyword ('f', "infix", INFIX);
	keyword ('h', "inherit", INHERIT);
	keyword ('l', "inline", INLINE);
	if ((pos+3<n) && (s[pos]=='t') && (s[pos+1]=='e') && (s[pos+2]=='r')) {
	  pos+=3;
	  keyword ('a', "interactive", INTERACTIVE);
	  keyword ('n', "intern", INTERN);
	}
      }
    }
    goto identifier;
  case 'j':
    if (pos<n) keyword ('o', "join", JOIN);
    goto identifier;
  case 'k':
    goto identifier;
  case 'l':
    if (((pos+1)<n) && (s[pos]=='o')) {
      pos++;
      keyword ('c', "locked", LOCKED);
      keyword ('o', "loop", LOOP);
    }
    goto identifier;
  case 'm':
    if (pos<n) keyword ('e', "method", METHOD);
    goto identifier;
  case 'n':
    goto identifier;
  case 'o':
    if (pos<n) {
      keyword ('r', "or", SEQOR);
      keyword ('u', "outline", OUTLINE);
    }
    goto identifier;
  case 'p':
    if ((pos+1)<n) {
      keyword ('a', "packed", PACKED);
      keyword ('o', "postfix", POSTFIX);
      if (s[pos]=='r') {
	pos++;
	if ((s[pos]=='e') && test_keyword("preamble")) {
	  while (pos+1<n) {
	    if ((s[pos]=='/') && (s[pos+1]=='{')) break;
	    pos++;
	  }
	  if (pos+1<n) pos += 2;
	  int start= pos;
	  while (pos+1<n) {
	    if ((s[pos]=='}') && (s[pos+1]=='/')) break;
	    pos++;
	  }
	  int end= pos;
	  if (pos+1<n) pos += 2;
	  string ss (s+ start, end- start);
	  parse_lval= generic (symbol (ss, org+ start, org+ end));
	  lex_accolade_mode= true;
	  return PREAMBLE;
	}
	keyword ('e', "prefix", PREFIX);
	keyword ('i', "private", PRIVATE);
	keyword ('o', "protected", PROTECTED);
      }
      keyword ('u', "public", PUBLIC);
    }
    goto identifier;
  case 'q':
    goto identifier;
  case 'r':
    if (pos<n) {
      keyword ('a', "raise", RAISE);
      keyword ('e', "return", RETURN);
    }
    goto identifier;
  case 's':
    if (pos<n) {
      keyword ('p', "split", SPLIT);
      keyword ('t', "step", STEP);
    }
    goto identifier;
  case 't':
    if (pos<n) {
      keyword ('h', "then", THEN);
      keyword ('r', "try", TRY);
    }
    goto identifier;
  case 'u':
    if (((pos+1)<n) && (s[pos]=='n')) {
      pos++;
      keyword ('p', "unpacked", UNPACKED);
      keyword ('t', "until", UNTIL);
    }
    goto identifier;
  case 'v':
    if (((pos+1)<n) && (s[pos]=='a')) {
      pos++;
      keyword ('l', "value", VALUE);
      keyword ('r', "variable", VARIABLE);
    }
    goto identifier;
  case 'w':
    if (pos<n) {
      keyword ('h', "while", WHILE);
      keyword ('i', "with", WITH);
    }
    goto identifier;
  case 'x':
    goto identifier;
  case 'y':
    if (pos<n) keyword ('i', "yield", YIELD);
    goto identifier;
  case 'z':
    goto identifier;
  case '{':
    produce (c);
  case '|':
    test ('|', OR);
    produce (WHERE);
  case '}':
    lex_accolade_mode= true;
    produce (c);
  case '~':
    produce (c);
  default:
    return lex ();

  identifier:
    while ((pos<n) && is_alpha (s[pos])) pos++;
    produce (IDENTIFIER);
  }

}

} // namespace mmx
